(* Example of a MPEG-2 video encoder as found in :
   Carloni and Sangiovanni-Vincentelli, Coping with LID in SOC Design, 
   IEEE Mico 2002

   Relay stations are the same as in :
   Casu and Macchiarulo, A New Approach to Latency Insensitive Design, DAC04
*)


let node ip_2_1 ((x::'a), (y::'a)) = x
let node ip_2_2 ((x::'a), (y::'a)) = (x ,x)

let node preprocessing x = x
let node frame_mem x = (x, x)
let node motion_est (x, y) = ip_2_2 (x, y)
let node motion_comp (x, y) = ip_2_2 (x, y)
let node dct x = x
let node quantizer (x, y) = ip_2_2 (x, y)
let node inverse_quantizer x = x
let node idct x = x
let node variable_length_code_enc (x, y) = ip_2_1 (x, y)
let node buff x = (x, x)
let node regulator x = x


let node mpeg input = output where
  rec a1 = preprocessing (buffer input)
  and a1' = merge 1(0) 4012 (delay a1)

  and (a2, a3) = frame_mem (buffer (delay a1'))
  and a2' = merge 1(0) 4012 (delay a2)
  and a3' = merge 1(0) 4012 (delay a3)

  and a4 = (buffer a2') - (buffer a13')
  and a4' = merge 1(0) 4012 (delay a4)

  and a5 = dct (buffer a4')
  and a5' = merge 1(0) 4012 (delay a5)

  and (a6, a16) = quantizer (buffer a5', 
			     buffer a20')
  and a6' = merge 1(0) 4012 (delay a6)
  and a16' = merge 1(0) 4012 (delay a16)

  and a7 = inverse_quantizer (buffer a6')
  and a7' = merge 1(0) 4012 (delay a7)

  and a8 = idct (buffer a7')
  and a8' = merge 1(0) 4012 (delay a8)

  and a9 = (buffer (delay a8')) + (buffer a12')
  and a9' = merge 1(0) 4012 (delay a9)

  and (a10, a11) = frame_mem (buffer a9')
  and a10' = merge 1(0) 4012 (delay a10)
  and a11' = merge 1(0) 4012 (delay a11)

  and (a12, a13) = motion_comp (buffer a10', 
				buffer (delay a14'))
  and a12' = merge 1(0) 4012 (delay a12)
  and a13' = merge 1(0) 4012 (delay a13)

  and (a14, a15) = motion_est (buffer (delay (delay a3')), 
			       buffer (delay a11'))
  and a14' = merge 1(0) 4012 (delay a14)
  and a15' = merge 1(0) 4012 (delay a15)

  and a17 = variable_length_code_enc (buffer a16', 
				      buffer (delay a15'))
  and a17' = merge 1(0) 4012 (delay a17)

  and (a18, a19) = buff (buffer a17')
  and a18' = merge 1(0) 4012 (delay a18)
  and a19' = merge 1(0) 4012 (delay a19)

  and a20 = regulator (buffer a19')
  and a20' = merge 1(0) 4012 (delay a20)

  and output =  a18'
